/*******************************************************************************
 * Licensed Materials - Property of IBM
 * "Restricted Materials of IBM"
 *
 *     Copyright IBM Corp. 2001, 2014 All Rights Reserved
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
package javax.rcm;

import java.util.concurrent.Executors;
import com.ibm.tenant.internal.TenantAccessFactory;

/**
 *	The background thread responsible for generating CPU throttling signal to threads.
 */
abstract class AbstractCPUSignalGenerator implements Runnable{

	private long interval = 1000;
	private boolean runThread = true;
	private Thread thread = null;
	private final Object syncObject = new Object(); /* Synchronization object used to wake-up and stop the underlying thread */

	/**
	 * The runnable for cpu throttling
	 */
	protected Runnable throttlingRunnable = CPUThrottlingRunnable.getInstance();

	/**
	 * Create the background thread
	 * @param interval  The interval of sending CPU throttling signal.
	 */
	protected AbstractCPUSignalGenerator(long interval){
		this.interval = interval;
	}

	/**
	 *
	 *
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		synchronized(syncObject) {
			while(runThread){
				try {
					try {
						signalCpuThrottling();
					} catch (OutOfMemoryError e) {
						// if we are in an OOM situation then just don't throttle until the next
						// interval
					}
					syncObject.wait(interval);
				} catch (InterruptedException e) {
					/* ignore spurious interruptions */
				}
			}
		}
	}

	/**
	 * Enumerate the threads that need CPU throttling, send signal.
	 */
	protected abstract void signalCpuThrottling();

	/**
	 * Terminate the daemon thread.
	 * @throws InterruptedException
	 * @returns true if the thread was successfully terminated, false otherwise
	 */
	synchronized boolean terminateDaemon(){
		final long timeoutInMilliseconds = 3 * 60 * 1000; /* three minutes */
		boolean terminated = false;

		synchronized(syncObject) {
			runThread = false;
			syncObject.notify();
		}

		try {
			thread.join(timeoutInMilliseconds);
			/* Ensure that the thread was successfully stopped */
			if(!thread.isAlive()) {
				terminated = true;
			}
		}catch(InterruptedException ie) {
		}

		return terminated;
	}

	/**
	 * Create and start the daemon thread.
	 */
	synchronized void createAndStartDaemon(){
		thread = newThread();
		thread.setDaemon(true);
		thread.setPriority((Thread.MAX_PRIORITY + Thread.NORM_PRIORITY)  / 2);

		final AbstractCPUSignalGenerator signalGenerator = this;

		/* Terminate 'throttling thread' during JVM shutdown. */
		Thread shutdownhook = new Thread(new Runnable(){
			public void run() {
				if (!signalGenerator.terminateDaemon()) {
					// R67 = Thread {0} is still alive after running the shutdown hooks.
					/* if the 'throttling thread' is still alive, throw a RuntimeException */
					throw new RuntimeException(com.ibm.oti.util.Msg.getString("R67", thread.getName())); //$NON-NLS-1$
				}
			}
		});

		/* add a shutdown hook to terminate the 'throttling thread' */
		Runtime.getRuntime().addShutdownHook(shutdownhook);

		synchronized(syncObject) {
			if (runThread) {
				/* We want to make sure the shutdown hook is added before
				 * the thread is started. And we only start the thread if
				 * stop() hasn't been called
				 */
				thread.start();
			}
		}
	}

	/**
	 * Creates a thread
	 * @return thread object
	 */
	protected abstract Thread newThread();

	/**
	 * Posts a request to execute an asynchronous event when the target thread
	 * reaches its next safe point.
	 * @param target The thread where the message should run.
	 * @param event The asynchronous event object.
	 * @return true if the request was posted, false otherwise.
	 * @throws NullPointerException if either the target thread or event are null.
	 */
	protected static native boolean signalAsyncEvent(Thread target, Runnable event);

}
