/*******************************************************************************
 * Licensed Materials - Property of IBM
 * "Restricted Materials of IBM"
 *
 *     Copyright IBM Corp. 2001, 2014 All Rights Reserved
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
package javax.rcm;

import java.lang.ref.PhantomReference;
import java.lang.ref.ReferenceQueue;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.ExecutorService;
import com.ibm.rcm.resolver.ResourceConsumerResolverCenter;
import com.ibm.rcm.resource.PlatformResourceAttributes;
import com.ibm.rcm.resource.RunningThreadsResourceAttributes;
import com.ibm.rcm.resource.config.PlatformConstraint;
import com.ibm.tenant.TenantContext;

/**
 * This class is a internal implementation of resource domain,
 * the ResourceDomain delegates its implementation to this class.
 */
class InternalResourceDomain {
    /**
     * A map to store all the active (not terminated) internal resource domains.
     */
    private static ConcurrentHashMap<ResourceAttributes, ConcurrentHashMap<String, InternalResourceDomain>> activeInternalDomains = new ConcurrentHashMap<ResourceAttributes, ConcurrentHashMap<String, InternalResourceDomain>>();

    /**
     * A map used to track the currently total available usage of managed
     * resources.
     */
    private static ConcurrentHashMap<ResourceAttributes, Long> totalFreeUsageOfBasicResources = new ConcurrentHashMap<ResourceAttributes, Long>();

    /**
     * A map to store usage consumed by domains which have terminated and
     * possible for being reclaimed by GC.
     */
    private static Map<ResourceAttributes, Long> totalUsageConsumedByTermiantedDomains = new HashMap<ResourceAttributes, Long>();

    /**
     * Pool size for notification executor.
     */
    static final int NOTIFICATION_POOL_SIZE = 5;

    /**
     * Notification executor to invoke notifications on resource domains
     * asynchronously.
     */
    private static ExecutorService notificationExec = null;

    /**
     * A map of resource attributes and phantom references.
     */
    private static ConcurrentHashMap<ResourceAttributes, Set<TerminatedDomainTracker>> terminatedInternalDomainTrackers = new ConcurrentHashMap<ResourceAttributes, Set<TerminatedDomainTracker>>();

    /**
     * Phantom reference queue.
     */
    private static ReferenceQueue<InternalResourceDomain> referenceQueue = new ReferenceQueue<InternalResourceDomain>();

    /**
     * The Phantom reference for this domain after terminated but still have
     * reservation on it. The reference is used to track the reservation on the
     * terminated domain, add return the reservation if the terminated domain
     * instance has been reclaimed by GC.
     */
    private TerminatedDomainTracker tracker;

    private final ResourceDomain domain;

    private final ResourceAttributes resourceAttributes;

    private final String name;

    /**
     * The current usage of this domain.
     */
    private volatile long currentUsage = 0;

    /**
     * The reservation of this resource domain.
     */
    private volatile long reservation = 0;

    /**
     * Set to be &quot;true&quot; after the domain has terminated.
     */
    private volatile boolean terminated = false;

    /**
     * All the user constraints applied to this domain.
     */
    private CopyOnWriteArrayList<Constraint> constraints = new CopyOnWriteArrayList<Constraint>();

    /**
     * All the notifications for this domain.
     */
    private CopyOnWriteArrayList<Notification> notifications = new CopyOnWriteArrayList<Notification>();

    /**
     * The type of consumer objects on this domain. Each domain supports only
     * one type of consumer object. And the type is determined after the first
     * resource consumer add to this domain.
     */
    private Class<?> consumerType;

    /**
     * Used for usage rate accounting.
     */
    private UsageRateRecorder usagaRateRecorder;

    /**
     * The delegation of this class, when resource to be managed is platform
     * resource.
     */
    private PlatformResourceDomainImpl platformImpl;

    /**
     * Sets the notification executor.
     * @param executor the executor to be set.
     */
    static void setNotificationExecutor(ExecutorService executor) {
    	notificationExec = executor;
    }
    /**
     * This class extends java.lang.ref.PhantomReference which tracks the
     * reservation on terminated resource domains. when the domain instance is
     * ready for being reclaimed by GC, the reservation will be released.
     *
     */
    private static class TerminatedDomainTracker extends
            PhantomReference<InternalResourceDomain> {
        private long reservation;

        TerminatedDomainTracker(InternalResourceDomain internalDomain,
                ReferenceQueue<InternalResourceDomain> queue) {
            super(internalDomain, queue);

            this.reservation = internalDomain.getReservation();

            /*
             * No need to track the active domains or domains have no
             * reservation.
             */
            if (!internalDomain.isTerminated() || 0 == reservation) {
                throw new AssertionError();
            }
        }

        long getReservation() {
            return this.reservation;
        }

        void setReservation(long reservation) {
            this.reservation = reservation;
        }
    }

    /**
     * This class is used for usage rate accounting. It keeps the history of
     * resource consumption of the last period from now.
     *
     */
    private static class UsageRateRecorder {
        // Keep the consume history record of the last period.
        Queue<ConsumeRecord> recordQueue = new ConcurrentLinkedQueue<ConsumeRecord>();

        final long interval;

        public UsageRateRecorder(long interval) {
            this.interval = interval;
        }

        synchronized void clearOutDateRecord() {
            long now = System.nanoTime();
            while (!recordQueue.isEmpty()
                    && (now - recordQueue.peek().time > interval)) {
                recordQueue.poll();
            }
        }

        /**
         * Records a resource consumption.
         *
         */
        void record(long quantity) {
            clearOutDateRecord();
            recordQueue.add(new ConsumeRecord(quantity));
        }

        /**
         * Get the usage rate (the usage of the last period from now).
         *
         */
        long getUsageRate() {
            clearOutDateRecord();
            long usage = 0;
            for (ConsumeRecord r : recordQueue) {
                usage += r.quantity;
            }
            return usage;
        }

        private static class ConsumeRecord {
            final long quantity;

            final long time;

            ConsumeRecord(long quantity) {
                this.time = System.nanoTime();
                this.quantity = quantity;
            }
        }
    }

    /**
     * Constructs an instance of <code>InternalResourceDomain</code> with
     * <code>ResourceDomain</code>.
     *
     * <strong>Note: </strong>If the resource is a
     * <code>PlatformResourceAttributes</code>, the instance of
     * PlatformResourceDomainImpl will be created
     * for delegating resource management to native.
     *
     * @param domain
     *            resource domain
     */
    @SuppressWarnings("boxing")
	InternalResourceDomain(ResourceDomain domain) {
        this.domain = domain;
        this.resourceAttributes = domain.getResourceAttributes();
        this.name = domain.getName();

        // No need to sync
        if (null == activeInternalDomains.get(resourceAttributes)) {
            activeInternalDomains.putIfAbsent(resourceAttributes,
                    new ConcurrentHashMap<String, InternalResourceDomain>());
        }

        // No need to sync
        ConcurrentHashMap<String, InternalResourceDomain> activeDomainsOfSameResource = activeInternalDomains
                .get(resourceAttributes);
        if (null != activeDomainsOfSameResource.putIfAbsent(name, this)) {
            // R28 = Domain "{0}" is not unique.
            throw new IllegalArgumentException(
            		com.ibm.oti.util.Msg.getString("R28", name)); //$NON-NLS-1$
        }

        if (isPlatformManaged()) {
        	platformImpl = new PlatformResourceDomainImpl(
                        (PlatformResourceAttributes) resourceAttributes, name);

        } else {
            synchronized (resourceAttributes) {
                /*
                 * Initializes available bandwidth of the resource on this
                 * domain.
                 */
                if (!resourceAttributes.isUnbounded()
                        && null == totalFreeUsageOfBasicResources
                                .get(resourceAttributes)) {
                    totalFreeUsageOfBasicResources.put(resourceAttributes,
                            resourceAttributes.getTotalQuantity());
                }
            }

            long period = resourceAttributes.getRateManagementPeriod();
            if (period > 0) {
                usagaRateRecorder = new UsageRateRecorder(period);
            }
        }
    }

    /**
     * Apply constraint to this domain.
     *
     * <strong>Note: </strong> if <code>resourceAttributes</code> on this domain
     * is {@link PlatformResourceAttributes}, setting bandwidth limit and
     * reservation is by adding a
     * {@link com.ibm.rcm.resource.config.PlatformConstraint platform
     * constraint} on the domain. Each domain has one platform constraint at
     * most. If there is a user constraint (not platform constraint) applied to
     * the domain, <em>native</em> will invoke the constraint callback into
     * <em>Java</em> at the interception point of resource consumption.
     *
     * @param constraint
     *            the constraint to be added
     *
     * @see javax.rcm.ResourceDomain#addConstraint(Constraint)
     */
    void addConstraint(Constraint constraint) {
        checkIfNull(constraint);
        checkTerminate();
        addConstraintImpl(constraint);
    }

    private void addConstraintImpl(Constraint constraint) {
        if (isPlatformManaged()) {
            if (constraint instanceof PlatformConstraint) {
                platformImpl
                        .addPlatformConstraint((PlatformConstraint) constraint);
            } else {
                synchronized (constraints) {
                    platformImpl.enableUserConstraints(this);
                    constraints.add(constraint);
                }
            }
        } else {
            // constraints only contains no-platformConstraints
            constraints.add(constraint);
        }
    }

    /**
     * Bound a consuming entity to this domain.
     *
     * @param consumerObj
     *            the consuming entity (see
     *            {@link javax.rcm.ResourceConsumer#getConsumer()})
     */
    void addConsumer(Object consumerObj) {
        // checkIfNull(consumer);
        checkTerminate();
        addConsumerImpl(consumerObj);
    }

    /*
     * TODO: There are also restrictions that the consumer type should support
     * disposable resource if the attributes is disposable.
     */
    private void addConsumerImpl(Object consumerObj) {
        if (consumerObj instanceof Thread) {
            /*
             * For thread consumer, active thread cannot be added to the
             * resource domain.
             */
            if (((Thread) consumerObj).isAlive()) {
                // R5E = The consumer has already started.
                throw new IllegalStateException(
                		com.ibm.oti.util.Msg.getString("R5E")); //$NON-NLS-1$
            }
        }

        synchronized (this) {
            checkConsumerType(consumerObj);

            if (isPlatformManaged()) {
                platformImpl.addConsumer(consumerObj);
            } else {
                if (consumerObj instanceof TenantContext) {
                    TenantContext tc = (TenantContext) consumerObj;

                    if (this.resourceAttributes instanceof RunningThreadsResourceAttributes) {
                        /*
                         * Register TenantContextListener for Tenants with
                         * RunningThreads limitation.
                         */
                    	com.ibm.tenant.internal.TenantAccessFactory.getAccess().addListener(tc, RCMRunningThreadsTenantListener.getInstance());
                    }

                }
            }

            if (null == consumerType) {
                /*
                 * Set the consumer type on this domain when addConsumer() is
                 * invoked the first time.
                 */
                setConsumerType(consumerObj);
            }
        }
    }

    private void checkConsumerType(Object consumerObj) {
        if (null == consumerType) {
            return;
        }

        if (null == consumerObj) {
            if (!consumerType.equals(Null.class)) {
                // R2B = Invalid consumer type, expect "{0}", but actually "{1}".
                throw new IllegalArgumentException(
                		com.ibm.oti.util.Msg.getString("R2B", new String[] { consumerType.getName(), //$NON-NLS-1$
                                "null" })); //$NON-NLS-1$
            }
        } else {
            if (!consumerType.isAssignableFrom(consumerObj.getClass())) {
                // R2B = Invalid consumer type, expect "{0}", but actually "{1}".
                throw new IllegalArgumentException(
                		com.ibm.oti.util.Msg.getString("R2B", new String[] { consumerType.getName(), //$NON-NLS-1$
                                consumerObj.getClass().getName() }));
            }
        }
    }

    private void setConsumerType(Object consumerObj) {
        if (null == consumerObj) {
            // In AUTO_PROCESS mode
            consumerType = Null.class;
        } else {
            /*
             * Consumer type on this domain is set to be the closet type
             * supported by the current resource consumer resolver.
             */
            for (Class<?> clazz : ResourceConsumerResolverCenter.getResolver()
                    .getConsumerTypes()) {
                if (consumerType == null) {
                    if (clazz.isAssignableFrom(consumerObj.getClass())) {
                        consumerType = clazz;
                    }
                } else {
                    if (consumerType.isAssignableFrom(clazz)) {
                        consumerType = clazz;
                    }
                }
            }
            assert (null != consumerType);
        }
    }

    /**
     * Add notification to this domain.
     *
     * <strong>Note: </strong> if <code>resourceAttributes</code> on this domain
     * is {@link PlatformResourceAttributes}, <em>native</em> will invoke the
     * notification callback into <em>Java</em> after each consumption request
     * arrives.
     *
     * @param notification
     *            the notification to be added
     *
     * @see javax.rcm.ResourceDomain#addNotification(Notification)
     */
    void addNotification(Notification notification) {
        checkIfNull(notification);
        checkTerminate();
        addNotificationImpl(notification);
    }

    private void addNotificationImpl(Notification notification) {
        synchronized (notifications) {
            if (isPlatformManaged()) {
                platformImpl.enableUserNotifications(this);
            }
            notifications.add(notification);
        }
    }

    /**
     * Invokes constraint callbacks associated to this domain to get their view
     * of what the resulting usage should be constrained to.
     *
     * <strong>Note: </strong> if <code>resourceAttributes</code> on this domain
     * is {@link PlatformResourceAttributes}, this method will be called from
     * <em>native</em> into <em>java</em> when a consumption request arrives and
     * there are user constraints applied to this platform resource domain.
     *
     * @param current
     *            resource usage at the time consumption is requested
     * @param proposed
     *            total usage that would result if the consumption request were
     *            to be granted in its entirety
     * @return the the resulting usage should be constrained to.
     */
    public long applyConstraints(final long current, final long proposed) {
        if (proposed < current) {
            // R41 = Trying to consume negative amount of resource.
            throw new IllegalArgumentException(
                    com.ibm.oti.util.Msg.getString("R41")); //$NON-NLS-1$
        }

        long result = proposed;
        long limited = 0;

        try {
            for (Constraint c : constraints) {
                limited = c.preConsume(domain, current, proposed);
                if (limited < current) {
                    result = current;
                    continue;
                }
                if (limited < result) {
                    result = limited;
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return result;
    }

    /**
     * Invokes notification callbacks on this domain to notify them the previous
     * usage before the consumption and the current usage.
     *
     * <strong>Note: </strong> if <code>resourceAttributes</code> on this domain
     * is {@link PlatformResourceAttributes}, this method will be called from
     * <em>native</em> into <em>java</em> when a consumption request arrives and
     * there are notifications added to this platform resource domain.
     *
     * @param perviousUsage
     *            usage at the time consumption was requested
     * @param finalUsage
     *            the total usage granted to the domain
     */
    void sendNotifications(final long perviousUsage,
            final long finalUsage) {
        try {
            Iterator<Notification> iter = notifications.iterator();
            while (iter.hasNext()) {
                final Notification notification = iter.next();
                notificationExec.execute(new Runnable() {
                    public void run() {
                        notification.postConsume(domain, perviousUsage,
                                finalUsage);
                    }
                });
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * @throws UnsupportedOperationException
     *             if <code>resourceAttributes</code> on this domain is
     *             {@link PlatformResourceAttributes}
     *
     * @see javax.rcm.ResourceDomain#consume(long, boolean)
     */
    long consume(long quantity, boolean allowPartialGrant) {
        if (isPlatformManaged()) {
            throw new UnsupportedOperationException();
        }

        checkTerminate();

        long granularity = resourceAttributes.getGranularity();
        if ((quantity % granularity) != 0) {
            quantity = ((long) Math.ceil(quantity * 1.0 / granularity))
                    * granularity;
        }

        long result = 0;
        long prevUsage;

        synchronized (this) {
            long proposed = applyConstraints(currentUsage, currentUsage
                    + quantity);

            if (!resourceAttributes.isUnbounded()) {
                if (proposed > reservation) {
                    synchronized (resourceAttributes) {
                        checkQueue();

                        @SuppressWarnings("boxing")
						long totalFree = totalFreeUsageOfBasicResources
                                .get(resourceAttributes);
                        long occupied = Math.max(reservation, currentUsage);
                        proposed = Math.min(proposed, totalFree + occupied);

                        if (proposed - currentUsage >= quantity
                                || allowPartialGrant) {
                            if (proposed > occupied) {
                                decreaseTotalRemain(proposed - occupied);
                            }
                        }
                    }
                }
            }

            result = Math.max(proposed - currentUsage, 0);
            if (result < quantity && !allowPartialGrant) {
                result = 0;
            }

            prevUsage = currentUsage;
            if (result > 0) {
                currentUsage += result;
            }
        }

        sendNotifications(prevUsage, currentUsage);

        if (null != usagaRateRecorder && result > 0) {
            usagaRateRecorder.record(result);
        }

        return result;
    }

    /**
     * @throws UnsupportedOperationException
     *             if <code>resourceAttributes</code> on this domain is
     *             {@link PlatformResourceAttributes}
     *
     * @see javax.rcm.ResourceDomain#relinquish(long)
     */
    long relinquish(long quantity) {
        if (isPlatformManaged()) {
            throw new UnsupportedOperationException();
        }

        checkTerminate();
        if (quantity < 0
                || (quantity % resourceAttributes.getGranularity()) != 0) {
            throw new IllegalArgumentException();
        }
        if (!resourceAttributes.isDisposable()) {
            // R2F = Resource is not disposable
            throw new IllegalStateException(com.ibm.oti.util.Msg.getString("R2F")); //$NON-NLS-1$
        }

        synchronized (this) {
            if (quantity > currentUsage) {
                // R30 = Attempting to relinquish more than is currently in use.
                throw new IllegalStateException(
                        com.ibm.oti.util.Msg.getString("R30")); //$NON-NLS-1$
            }

            if (currentUsage > reservation) {
                synchronized (resourceAttributes) {
                    increaseTotalRemain(currentUsage
                            - Math.max(currentUsage - quantity, reservation));
                }
            }
            currentUsage -= quantity;
            sendNotifications(currentUsage + quantity, currentUsage);
        }
        return quantity;
    }

    /**
     * Remove the constraint from this domain.
     *
     * <strong>Note: </strong> if <code>resourceAttributes</code> on this domain
     * is {@link PlatformResourceAttributes} and the constraint to be removed is
     * an instance of PlatformConstraint, then management of rate of consumption
     * provided by native will be disabled.
     *
     * @see javax.rcm.ResourceDomain#removeConstraint(javax.rcm.Constraint)
     */
    void removeConstraint(Constraint constraint) {
        checkTerminate();
        checkIfNull(constraint);
        removeConstraintImpl(constraint);
    }

    private void removeConstraintImpl(Constraint constraint) {
        if (isPlatformManaged()) {
            if (constraint instanceof PlatformConstraint) {
                platformImpl
                        .removePlatformConstraint((PlatformConstraint) constraint);
            } else {
                synchronized (constraints) {
                    constraints.remove(constraint);
                    if (constraints.isEmpty()) {
                        platformImpl.disableUserConstraints();
                    }
                }
            }
        } else {
            constraints.remove(constraint);
        }
    }

    /**
     * Remove the consuming entity from this domain.
     *
     * @param consumerObj
     *            the consuming entity to be removed from this domain
     *
     * @see javax.rcm.ResourceDomain#removeConsumer(javax.rcm.ResourceConsumer)
     */
    void removeConsumer(Object consumerObj) {
        // checkIfNull(resourceConsumer);
        checkTerminate();
        removeConsumerImpl(/* resourceConsumer, */consumerObj);
    }

    private void removeConsumerImpl(Object consumerObj) {
        synchronized (this) {
            checkConsumerType(consumerObj);

            /*
             * If the resource is disposable, getUsage() must be zero when this
             * method is invoked; otherwise, an exception will be thrown.
             */
            if (resourceAttributes.isDisposable() && getUsage() > 0) {
                // R32 = Can not terminate the resource domain now as its resource is disposable and the current usage is not zero.
                throw new IllegalStateException(
                        com.ibm.oti.util.Msg.getString("R32")); //$NON-NLS-1$
            }

            if (isPlatformManaged()) {
                platformImpl.removeConsumer(consumerObj);
            } else {
            	 if (consumerObj instanceof TenantContext) {
                     TenantContext tc = (TenantContext) consumerObj;

                     /*
                      * Unregister the TenantContext listeners.
                      */
                     if (this.resourceAttributes instanceof RunningThreadsResourceAttributes) {
                    	 com.ibm.tenant.internal.TenantAccessFactory.getAccess().removeListener(tc, RCMRunningThreadsTenantListener.getInstance());
                     }
                 }
            }

        }
    }

    /**
     * Remove the notification from this domain
     *
     * @see javax.rcm.ResourceDomain#removeNotification(javax.rcm.Notification)
     */
    void removeNotification(Notification notification) {
        checkIfNull(notification);
        checkTerminate();
        removeNotificationImpl(notification);
    }

    private void removeNotificationImpl(Notification notification) {
        synchronized (notifications) {
            notifications.remove(notification);
            if (isPlatformManaged()) {
                if (notifications.isEmpty()) {
                    platformImpl.disableUserNotifications();
                }
            }
        }
    }

    /**
     * @throws UnsupportedOperationException
     *             if <code>resourceAttributes</code> on this domain is
     *             {@link PlatformResourceAttributes}
     *
     * @see javax.rcm.ResourceDomain#setReservation(long)
     */
    void setReservation(long newValue) {
        if (isPlatformManaged()) {
            throw new UnsupportedOperationException();
        }

        if (newValue < 0) {
            throw new IllegalArgumentException();
        }

        checkIfReservable();

        synchronized (this) {
            if (terminated) {
                if (newValue > reservation) {
                    // R33 = Reservation can only be decreased on terminated resource domain.
                    throw new IllegalStateException(
                            com.ibm.oti.util.Msg.getString("R33")); //$NON-NLS-1$
                }

                if (null != tracker) {
                    if (0 == newValue) {
                        /*
                         * If the terminated domain has no reservation, there's
                         * no need to track it any more.
                         */
                        terminatedInternalDomainTrackers
                                .get(resourceAttributes).remove(tracker);
                        tracker = null;
                    } else {
                        tracker.setReservation(newValue);
                    }
                }
            }

            long occupied = Math.max(this.reservation, this.currentUsage);
            long additional = newValue - occupied;

            if (0 != additional) {
                synchronized (resourceAttributes) {
                    checkQueue();

                    @SuppressWarnings("boxing")
					long totalRemain = totalFreeUsageOfBasicResources
                            .get(resourceAttributes);
                    if (additional > totalRemain) {
                        // R34 = Reservation cannot be satisfied.
                        throw new ResourceException(
                                com.ibm.oti.util.Msg.getString("R34")); //$NON-NLS-1$
                    }
                    if (additional > 0) {
                        decreaseTotalRemain(additional);
                    } else if (reservation > currentUsage
                            && reservation > newValue) {
                        increaseTotalRemain(reservation - newValue);
                    }
                }
            }

            this.reservation = newValue;
        }
    }

    /**
     * Terminate this resource domain.
     *
     * After the domain terminated, only query methods on this domain can be
     * invoked.
     *
     * @see javax.rcm.ResourceDomain#terminate()
     */
    @SuppressWarnings("boxing")
	void terminate() {
        synchronized (this) {
            if (terminated) {
                return; // If the domain has already terminated, just return
            }

            if (isPlatformManaged()) {
                platformImpl.terminate();
            } else {
                usagaRateRecorder = null;
            }

            this.terminated = true;

            /*
             * Remove the terminated resource domain from the list, so that it
             * won't permit GC collecting the domain instance.
             */
            synchronized (resourceAttributes) {
                activeInternalDomains.get(resourceAttributes).remove(name);

                if (!isPlatformManaged()) {
                    if (currentUsage > 0) {
                        Long usageOfTerminatedDomains = totalUsageConsumedByTermiantedDomains
                                .get(resourceAttributes);
                        if (null == usageOfTerminatedDomains) {
                            totalUsageConsumedByTermiantedDomains.put(
                                    resourceAttributes, currentUsage);
                        } else {
                            totalUsageConsumedByTermiantedDomains.put(
                                    resourceAttributes,
                                    usageOfTerminatedDomains.longValue()
                                            + currentUsage);
                        }
                    }

                    if (reservation > 0) {
                        tracker = new TerminatedDomainTracker(this,
                                referenceQueue);
                        terminatedInternalDomainTrackers
                                .putIfAbsent(
                                        resourceAttributes,
                                        Collections
                                                .newSetFromMap(new ConcurrentHashMap<TerminatedDomainTracker, Boolean>()));
                        terminatedInternalDomainTrackers
                                .get(resourceAttributes).add(tracker);
                    }
                }
            }
        }
    }

    String getName() {
        return this.name;
    }

    ResourceAttributes getResourceAttributes() {
        return this.resourceAttributes;
    }

    /**
     * Get the usage reserved by all the domains of the resource.
     *
     * @see javax.rcm.ResourceDomain#getTotalReservation()
     */
    long getTotalReservation() {
        checkIfReservable();
        assert (!isPlatformManaged());

        long total = 0;
        for (InternalResourceDomain internal : getActiveDomainsOfTheResource(resourceAttributes)) {
            total += internal.reservation;
        }
        if (terminatedInternalDomainTrackers.containsKey(resourceAttributes)) {
            for (TerminatedDomainTracker theTracker : terminatedInternalDomainTrackers
                    .get(resourceAttributes)) {
                total += theTracker.getReservation();
            }
        }
        return total;
    }

    /**
     * Get the usage reserved by this domain.
     *
     * @see javax.rcm.ResourceDomain#getReservation()
     */
    long getReservation() {
        checkIfReservable();
        return this.reservation;
    }

    /**
     * Get the usage consumed by this domain.
     *
     * If the resource is platform managed, usage accounting result is returned
     * from <em>native</em>.
     *
     * @see javax.rcm.ResourceDomain#getUsage()
     */
    long getUsage() {
        if (isPlatformManaged()) {
            return platformImpl.getUsage();
        } else {
            if (this.terminated) {
                return this.currentUsage;
            }
            return this.currentUsage;
        }
    }

    /**
     * Returns the rate of usage of the resource by this domain.
     *
     * @see javax.rcm.ResourceDomain#getUsageRate()
     */
    long getUsageRate() {
        if (resourceAttributes.getRateManagementPeriod() <= 0) {
            // R2E = Resource rates are not managed: rate management period on "{0}" is not set.
            throw new IllegalStateException(
            		com.ibm.oti.util.Msg.getString("R2E", resourceAttributes.getName())); //$NON-NLS-1$
        }

        if (isPlatformManaged()) {
            return platformImpl.getUsageRate();
        } else {
            if (terminated) {
                // If the domain is terminated, usage rate is 0.
                return 0;
            }
            assert (usagaRateRecorder != null);
            return usagaRateRecorder.getUsageRate();
        }
    }

    /**
     * Returns the usage consumed by all domains for this resource.
     *
     * @see javax.rcm.ResourceDomain#getTotalUsage()
     */
    @SuppressWarnings("boxing")
	long getTotalUsage() {
        long total = 0;
        for (InternalResourceDomain internal : getActiveDomainsOfTheResource(resourceAttributes)) {
            total += internal.getUsage();
        }
        if (isPlatformManaged()) {
            total += PlatformResourceDomainImpl
                    .getTotalUsageOfTermiantedDomains((PlatformResourceAttributes) resourceAttributes);
        } else {
            if (totalUsageConsumedByTermiantedDomains
                    .containsKey(resourceAttributes)) {
                total += totalUsageConsumedByTermiantedDomains
                        .get(resourceAttributes);
            }
        }
        return total;
    }

    /**
     * Returns the rate of usage by all domains for this resource.
     *
     * @see javax.rcm.ResourceDomain#getTotalUsageRate()
     */
    long getTotalUsageRate() {
        long total = 0;
        for (InternalResourceDomain internal : getActiveDomainsOfTheResource(resourceAttributes)) {
            if (null != internal) {
                total += internal.getUsageRate();
            }
        }
        return total;
    }

    /**
     * Determine whether this domain has terminated or not.
     *
     * @see javax.rcm.ResourceDomain#isTerminated()
     */
    boolean isTerminated() {
        return terminated;
    }

    /**
     * Checks if the domain has been terminated.
     */
    private void checkTerminate() {
        if (terminated) {
            // R36 = Resource domain is terminated.
            throw new IllegalStateException(com.ibm.oti.util.Msg.getString("R36", //$NON-NLS-1$
                    name));
        }
    }

    /**
     * Checks if the target object is null.
     */
    private void checkIfNull(Object obj) {
        if (null == obj) {
            throw new NullPointerException();
        }
    }

    /**
     * Checks if the resource is reservable
     */
    private void checkIfReservable() {
        if (!resourceAttributes.isReservable()) {
            // R2D = "{0}" is not reservable.
            throw new IllegalStateException(
            		com.ibm.oti.util.Msg.getString("R2D", resourceAttributes.getName())); //$NON-NLS-1$
        }
    }

    public ResourceDomain getResourceDomain() {
        return this.domain;
    }

    /**
     * Returns the active resource domains of resource
     * @param attributes the resource in question.
     * @return
     */
    private Collection<InternalResourceDomain> getActiveDomainsOfTheResource(
            ResourceAttributes attributes) {
        Map<String, InternalResourceDomain> domainMap = activeInternalDomains
                .get(attributes);
        if (null != domainMap) {
            return domainMap.values();
        }
        return null;
    }

    /**
     *  Increases total free usage of resource.
     * @param quantity the amount to be increased.
     */
    @SuppressWarnings("boxing")
	private void increaseTotalRemain(long quantity) {
		long totalRemain = totalFreeUsageOfBasicResources
                .get(resourceAttributes);
        totalRemain += quantity;
        totalFreeUsageOfBasicResources.put(resourceAttributes, totalRemain);
    }

    /**
     * Decreases total free usage of resource.
     * @param quantity the amount to be decreased.
     */
    @SuppressWarnings("boxing")
	private void decreaseTotalRemain(long quantity) {
        long totalRemain = totalFreeUsageOfBasicResources
                .get(resourceAttributes);
        totalRemain -= quantity;
        totalFreeUsageOfBasicResources.put(resourceAttributes, totalRemain);
    }

    /**
     * Answers if the resource is platform managed.
     * @return true if it is.
     */
    private boolean isPlatformManaged() {
        return (resourceAttributes instanceof PlatformResourceAttributes);
    }

    /**
     * Updates total free usage of resource.
     */
    @SuppressWarnings("boxing")
	private void checkQueue() {
        TerminatedDomainTracker toRemove = null;
        while (null != (toRemove = (TerminatedDomainTracker) referenceQueue
                .poll())) {
            totalFreeUsageOfBasicResources.put(resourceAttributes,
                    totalFreeUsageOfBasicResources.get(resourceAttributes)
                            + toRemove.getReservation());
            terminatedInternalDomainTrackers.get(resourceAttributes).remove(
                    toRemove);
            toRemove.clear();
        }
    }

    /**
     * Dummy class for the type of JVM consumer.
     */
    class Null {
    };
}
