/*******************************************************************************
 * Licensed Materials - Property of IBM
 * "Restricted Materials of IBM"
 *
 *     Copyright IBM Corp. 2001, 2014 All Rights Reserved
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
package javax.rcm;

import com.ibm.rcm.resource.config.PlatformConstraint;

/**
 * This class supports handling consumption management of platform resource
 * by underlying layer.
 *
 */
class PlatformResourceDomainHandler {

    /*
     * private constructor
     */
    private PlatformResourceDomainHandler() {
    };

    private static PlatformResourceDomainHandler instance = new PlatformResourceDomainHandler();

    /*
     * Singleton.
     *
     * @return the single instance of this class.
     */
    static PlatformResourceDomainHandler getInstance() {
        return instance;
    }

    /**
     * Creates a <strong>virtual resource container</strong> which is a
     * structure defined and implemented in native for providing resource
     * management and accounting for platform resource.
     *
     * This method is invoked when a resource domain is created for controlling
     * the consumption of a platform resource.
     *
     * @param resourceHandle
     *            The pointer to the platform resource structure in native.
     * @param domainName
     *            The name of the resource domain
     * @return the pointer to the virtual resource container
     *
     */
    native long createVirtualResourceContainer(long resourceHandle,
            String domainName);

    /**
     * Destroys the virtual resource container.
     *
     * This method is invoked after a resource domain controlling a platform
     * resource is terminated.
     *
     * @param vrcHandle
     *            The pointer to the virtual resource container
     * @return 0 for success, otherwise for failure
     */
    native int destroyVirtualResourceContainer(long vrcHandle);

    /**
     * Enables the control of the rate of consumption for the managed platform
     * resource on the virtual resource container.
     *
     * @param vrcHandle
     *            The pointer to the virtual resource container
     * @param platformConstraint
     *            The platform constraint which expresses the policy of
     *            consumption rate management.
     * @return 0 for success, otherwise for failure
     */
    native int enableUsageRateControl(long vrcHandle,
            PlatformConstraint platformConstraint);

    /**
     * Updates the policy of consumption rate management.
     *
     * @param vrcHandle
     *            The pointer to the virtual resource container
     * @param platformConstraint
     *            The platform constraint which expresses the policy of
     *            consumption rate management.
     * @return 0 for success, otherwise for failure
     */
    native int updateUsageRateControl(long vrcHandle,
            PlatformConstraint platformConstraint);

    /**
     * Disables consumption rate management on the virtual resource container.
     *
     * @param vrcHandle
     *            The pointer to the virtual resource container
     * @return 0 for success, otherwise for failure
     */
    native int disableUsageRateControl(long vrcHandle);

    /**
     * Enables a call from native into Java to execute the user constraints
     * associated to the domain in each time native intercepting a consumption
     * request.
     *
     * This method is invoked when detecting at least one user constraint has
     * applied to the domain.
     *
     * @param vrcHandle
     *            The pointer to the virtual resource container
     * @param domain
     *            The resource domain instance.
     * @return 0 for success, otherwise for failure
     */
    native int enableUserConstraints(long vrcHandle, ResourceDomain domain);

    /**
     * Disables invoking user constraint callbacks from native to Java.
     *
     * This method is invoked when there's no user constraint associate to the
     * domain to avoid unnecessary calls from native into Java.
     *
     * @param vrcHandle
     *            The pointer to the virtual resource container
     * @return 0 for success, otherwise for failure
     */
    native int disableUserConstraints(long vrcHandle);

    /**
     * Enables a call from native into Java to execute the notifications
     * associated to the domain each time after a consumption request arrives.
     *
     * This method is invoked when detecting at least one notification has
     * associated to the domain.
     *
     * @param vrcHandle
     *            The pointer to the virtual resource container
     * @param domain
     *            The resource domain instance
     * @return 0 for success, otherwise for failure
     */
    native int enableUserNotifications(long vrcHandle, ResourceDomain domain);

    /**
     * Disables invoking notification callbacks.
     *
     * This method is invoked when there's no notification associate to the
     * domain to avoid unnecessary calls from native into Java.
     *
     * @param vrcHandle
     *            The pointer to the virtual resource container
     * @return 0 for success, otherwise for failure
     */
    native int disableUserNotifications(long vrcHandle);

    /**
     * Bound a consuming entity to the virtual resource container.
     *
     * @param vrcHandle
     *            The pointer to the virtual resource container
     * @param consumerObj
     *            The consuming entity
     * @return 0 for success, otherwise for failure
     */
    native int addConsumer(long vrcHandle, Object consumerObj);

    /**
     * Unbound the consuming entity from the virtual resource container.
     *
     * @param vrcHandle
     *            The pointer to the virtual resource container
     * @param consumerObj
     *            The consuming entity
     * @return 0 for success, otherwise for failure
     */
    native int removeConsumer(long vrcHandle, Object consumerObj);

    /**
     * Gets the usage of the platform resource by consuming entities bounded to
     * the virtual resource container.
     *
     * @param vrcHandle
     *            The pointer to the virtual resource container
     * @param resHandle
     *            The pointer to the platform resource whose usage accounting is
     *            provided by the virtual resource container
     * @return the usage of the platform resource by consuming entities bound to
     *         the virtual resource container
     * @throws IllegalArgumentException
     *             if <code>resHandle</code> is not recognized by the virtual
     *             resource container.
     */
    native long getUsage(long vrcHandle, long resHandle);

    /**
     * Gets the rate of usage of the platform resource by the virtual resource
     * container.
     *
     * @param vrcHandle
     *            The pointer to the virtual resource container
     * @param resHandle
     *            The pointer to the platform resource whose usage rate
     *            accounting is provided by the virtual resource container
     * @return the usage rate of the resource
     * @throws IllegalArgumentException
     *             if <code>resHandle</code> is not recognized by the virtual
     *             resource container.
     */
    native long getUsageRate(long vrcHandle, long resHandle);

}
