/*******************************************************************************
 * Licensed Materials - Property of IBM
 * "Restricted Materials of IBM"
 *
 *     Copyright IBM Corp. 2001, 2014 All Rights Reserved
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
package javax.rcm;

import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import com.ibm.rcm.resource.CpuTimeResourceAttributes;
import com.ibm.rcm.resource.PlatformResourceAttributes;
import com.ibm.rcm.resource.config.PlatformConstraint;
import com.ibm.tenant.TenantContext;

/**
 * This class provides consumption management of platform resources. If consumer
 * type is {@link com.ibm.tenant.TenantContext}, this class will register to be
 * a listener to tenant attach/detach event to ensure the request consumption by
 * all the threads attached by the tenant has managed.
 */
class PlatformResourceDomainImpl  {

    /**
     * The platform resource domain handler for handling consumption management
     * of platform resource by native.
     */
    volatile PlatformResourceDomainHandler handler;

    /**
     * The handle of virtual resource container (see
     * {@link PlatformResourceDomainHandler#createVirtualResourceContainer(long, String)
     * createVirtualResourceContainer(long, String)} )
     */
    long vrc;

    /**
     * A map to calculate total usage consumed by terminated platform resource
     * domains.
     */
    static Map<ResourceAttributes, Long> totalUsageConsumedByTermiantedPlatformDomains = new ConcurrentHashMap<ResourceAttributes, Long>();

    /**
     * A map to store available bandwidth of platform resources.
     */
    static ConcurrentHashMap<PlatformResourceAttributes, Long> totalFreeBandwidthOfPlatformResources = new ConcurrentHashMap<PlatformResourceAttributes, Long>();

    /**
     * Express the policy of controlling the rate of usage of this platform
     * resource.
     */
    PlatformConstraint platformConstraint;

    /**
     * Set to TRUE when there's user constraints associated to this domain.
     */
    boolean enableUserConstraints;

    /**
     * Set to TRUE when there's notifications associated to this domain
     */
    boolean enableUserNotifications;

    /**
     * The platform resource controlled by this domain
     */
    final PlatformResourceAttributes pattr;

    /**
     * The name of this domain
     */
    final String domainName;

    /**
     * usage consumed by this domain after terminated.
     */
    long usage;

    /**
     * Constructs an instance of <code>PlatformResourceDomainImpl</code> with
     * platform resource and unique name.
     *
     * @param pattr
     *            the platform resource
     * @param domainName
     *            the domain name
     */
    public PlatformResourceDomainImpl(PlatformResourceAttributes pattr,
            String domainName) {
        this.pattr = pattr;
        this.domainName = domainName;
        this.handler = PlatformResourceDomainHandler.getInstance();
        this.vrc = handler.createVirtualResourceContainer(pattr.getHandle(),
                domainName);
        // Invalid value
        if (vrc == 0) {
            // R13 = Fail to create the virtual resource container for resource domain:\ {0}.
            throw new ResourceException(
            		com.ibm.oti.util.Msg.getString("R13", domainName)); //$NON-NLS-1$
        }
    }

    /**
     * Associate a <code>PlatformConstraint</code> to this domain.
     *
     * @param pc
     *            the platform constraint to be added to this domain.
     */
    public void addPlatformConstraint(PlatformConstraint pc) {
        validate(pc);

        synchronized (this) {
            if (null == this.platformConstraint) {
                accquireBWReservation(pc);
                enableUsageRateControl(pc);
                this.platformConstraint = pc;
            } else {
                /*
                 * If the domain already has an existing platformConstraint,
                 * throw exception.
                 */
                // R2A = Resource domain "{0}" has already has a platform constraint.
                throw new ResourceException(com.ibm.oti.util.Msg.getString("R2A", //$NON-NLS-1$
                        domainName));
            }
        }
    }

    public void removePlatformConstraint(PlatformConstraint pc) {
        synchronized (this) {
            if (null != this.platformConstraint
                    && this.platformConstraint.equals(pc)) {
                disableUsageRateControl();
                releaseBWReservation();
                this.platformConstraint = null;
            }
        }
    }

    @SuppressWarnings("boxing")
	boolean decreaseFreeBandwidth(PlatformResourceAttributes thePattr,
            long reservation) {
        Long remain = totalFreeBandwidthOfPlatformResources.get(thePattr);
        if (null == remain) {
            remain = thePattr.getMaxBandwidth();
        }
        remain -= reservation;
        if (remain < 0) {
            return false;
        }
        totalFreeBandwidthOfPlatformResources.put(thePattr, remain);
        return true;
    }

    @SuppressWarnings("boxing")
	void increaseFreeBandwidth(PlatformResourceAttributes thePattr,
            long reservation) {
		long remain = totalFreeBandwidthOfPlatformResources.get(thePattr);
        remain += reservation;
        if (remain > thePattr.getMaxBandwidth()) {
            throw new AssertionError();
        }
        totalFreeBandwidthOfPlatformResources.put(thePattr, remain);
    }

    void accquireBWReservation(PlatformConstraint pc) {
        long reservation = pc.getBandwidthReservation(pattr.getHandle());
        synchronized (pattr) {
            if (!decreaseFreeBandwidth(pattr, reservation)) {
                // R29 = Fail to add platform constraint on resource domain: "{0}". There's no enough bandwidth for platform resource "{1}".
                throw new ResourceException(
                		com.ibm.oti.util.Msg.getString("R29", new String[] { domainName, pattr.getName() })); //$NON-NLS-1$
            }
        }
    }

    void releaseBWReservation() {
        long reservation = platformConstraint.getBandwidthReservation(pattr
                .getHandle());
        synchronized (pattr) {
            increaseFreeBandwidth(pattr, reservation);
        }
    }

    void validate(PlatformConstraint thePlatformConstraint) {
        if (!thePlatformConstraint.getResource().equals(pattr)) {
            // R3C = Invalid platform constraint. Expect the resource to be {0}, but {1}.
            throw new IllegalArgumentException(
            		com.ibm.oti.util.Msg.getString("R3C", new Object[] { pattr, thePlatformConstraint.getResource() })); //$NON-NLS-1$
        }

        long reservation = thePlatformConstraint.getBandwidthReservation(pattr
                .getHandle());
        long limit = thePlatformConstraint.getBandwidthLimit(pattr.getHandle());
        if (reservation > limit) {
            // R4A = Invalid configure for resource "{0}": bandwidth reservation should not greater than bandwidth limit.
            throw new IllegalArgumentException(
            		com.ibm.oti.util.Msg.getString("R4A", pattr.getName())); //$NON-NLS-1$
        }
    }

    private static final int TC_ADD_HTB_CLASS_FAILED = 49996;

    private static final int TC_REMOVE_HTB_CLASS_FAILED = 49997;

    void enableUsageRateControl(PlatformConstraint pc) {
        int rt = handler.enableUsageRateControl(vrc, pc);
        if (rt != 0) {
            if (rt == TC_ADD_HTB_CLASS_FAILED) {
// R14 = Fail to enable usage rate control for resource domain:\ {0}.
                // R5B = Add HTB class to traffic control failed.
                throw new ResourceException(com.ibm.oti.util.Msg.getString("R14", //$NON-NLS-1$
                        domainName) + com.ibm.oti.util.Msg.getString("R5B")); //$NON-NLS-1$
            } else {
                // R14 = Fail to enable usage rate control for resource domain:\ {0}.
                throw new ResourceException(com.ibm.oti.util.Msg.getString("R14", //$NON-NLS-1$
                        domainName));
            }
        }
    }

    void updateUsageRateControl(PlatformConstraint pc) {
        int rt = handler.updateUsageRateControl(vrc, pc);
        if (rt != 0) {
            // R15 = Fail to update usage rate control for resource domain:\ {0}.
            throw new ResourceException(
            		com.ibm.oti.util.Msg.getString("R15", domainName)); //$NON-NLS-1$
        }
    }

    void disableUsageRateControl() {
        int rt = handler.disableUsageRateControl(vrc);
        if (rt != 0) {
            // R16 = Fail to disable usage rate control for resource domain:\ {0}.
            if (rt == TC_REMOVE_HTB_CLASS_FAILED) {
                // R5C = Remove HTB class from traffic control failed.
                throw new ResourceException(
                		com.ibm.oti.util.Msg.getString("R16", domainName) //$NON-NLS-1$
                        + com.ibm.oti.util.Msg.getString("R5C")); //$NON-NLS-1$
            }
            throw new ResourceException(
            		com.ibm.oti.util.Msg.getString("R16", domainName)); //$NON-NLS-1$
        }
    }

    /**
     * Adds a resource consumer to this platform resource domain.
     *
     * @param consumerObj
     *            the consumer object
     */
    public void addConsumer(Object consumerObj) {
        int rt = handler.addConsumer(vrc, consumerObj);
        if (rt != 0) {
            // R17 = Fail to add consumer on resource domain:\ {0}.
            throw new ResourceException(
            		com.ibm.oti.util.Msg.getString("R17", domainName)); //$NON-NLS-1$
        }

        if (consumerObj instanceof TenantContext) {
            TenantContext tc = (TenantContext) consumerObj;
            if (pattr instanceof CpuTimeResourceAttributes ) {
                /*
                 * Register TenantContextListener for Tenants with JVM CPU
                 * Throttling.
                 */
            	com.ibm.tenant.internal.TenantAccessFactory.getAccess().addListener(tc, CPUThrottlingRunnable.getInstance());
            }
        }
    }

    /**
     * Removes the resource consumer from this platform resource domain.
     *
     * @param consumerObj
     *            the consumer object
     */
    public void removeConsumer(Object consumerObj) {
        int rt = handler.removeConsumer(vrc, consumerObj);
        if (rt != 0) {
            // R18 = Fail to remove consumer on resource domain:\ {0}.
            throw new ResourceException(
            		com.ibm.oti.util.Msg.getString("R18", domainName)); //$NON-NLS-1$
        }

        if (consumerObj instanceof TenantContext) {
            TenantContext tc = (TenantContext) consumerObj;
            if (pattr instanceof CpuTimeResourceAttributes ) {
            	com.ibm.tenant.internal.TenantAccessFactory.getAccess().removeListener(tc, CPUThrottlingRunnable.getInstance());
            }
        }
    }

    /**
     * Enables user constraints on this platform resource domain, so that they
     * will be invoked whenever there's consumption request on this platform
     * resource.
     * @param internalDomain the wrapped internal resource domain.
     */
    public synchronized void enableUserConstraints(
            InternalResourceDomain internalDomain) {
        if (!enableUserConstraints) {
            int rt = handler.enableUserConstraints(vrc,
                    internalDomain.getResourceDomain());
            if (rt != 0) {
                // R1B = Fail to enable user constraints on resource domain:\ {0}.
                throw new ResourceException(
                		com.ibm.oti.util.Msg.getString("R1B", domainName)); //$NON-NLS-1$
            }
            enableUserConstraints = true;
        }
    }

    /**
     * Disables invoking user constraints from <em>native</em> to <em>Java</em>
     * on this domain.
     */
    public synchronized void disableUserConstraints() {
        if (enableUserConstraints) {
            int rt = handler.disableUserConstraints(vrc);
            if (rt != 0) {
                // R1C = Fail to disable user constraints on resource domain:\ {0}.
                throw new ResourceException(
                		com.ibm.oti.util.Msg.getString("R1C", domainName)); //$NON-NLS-1$
            }
            enableUserConstraints = false;
        }
    }

    /**
     * Enables user notifications on this platform resource domain, so that they
     * will be invoked whenever there's consumption request on this platform
     * resource.
     * @param internalDomain the wrapped internal resource domain.
     */
    public synchronized void enableUserNotifications(
            InternalResourceDomain internalDomain) {
        if (!enableUserNotifications) {
            int rt = handler.enableUserNotifications(vrc,
                    internalDomain.getResourceDomain());
            if (rt != 0) {
                // R1D = Fail to enable user notifications on resource domain:\ {0}.
                throw new ResourceException(
                		com.ibm.oti.util.Msg.getString("R1D", domainName)); //$NON-NLS-1$
            }
            enableUserNotifications = true;
        }
    }

    /**
     * Disables invoking notifications from <em>native</em> to <em>Java</em> on
     * this domain.
     */
    public synchronized void disableUserNotifications() {
        if (enableUserNotifications) {
            int rt = handler.disableUserNotifications(vrc);
            if (rt != 0) {
                // R1E = Fail to disable user notifications on resource domain:\ {0}.
                throw new ResourceException(
                		com.ibm.oti.util.Msg.getString("R1E", domainName)); //$NON-NLS-1$
            }
            enableUserNotifications = false;
        }
    }

    /**
     * Gets the rate of usage of the resource by this domain.
     *
     * @return the rate of usage of the resource by this domain
     *
     */
    public long getUsageRate() {
        if (isTermiante()) {
            return 0;
        }
        return getUsageRateImpl(pattr);
    }

    long getUsageRateImpl(PlatformResourceAttributes thePattr) {
        if (!thePattr.isUsageRateMonitoringEnabled()) {
            // R62 = Resource rates are not managed: the attribute 'usageRateMonitoringEnabled' of "{0}" is false.
            throw new IllegalStateException(
            		com.ibm.oti.util.Msg.getString("R62", thePattr.getName())); //$NON-NLS-1$
        }

        long result = handler.getUsageRate(vrc, thePattr.getHandle());
        if (result < 0) {
            throw new UnsupportedOperationException();
        }
        return result;
    }

    /**
     * Gets resource usage consumed by this domain.
     *
     * @return the usage consumed by this domain.
     */
    public long getUsage() {
        if (isTermiante()) {
            return this.usage;
        }
        return getUsageImpl(pattr);
    }

    long getUsageImpl(PlatformResourceAttributes thePattr) {
        long result = handler.getUsage(vrc, thePattr.getHandle());
        if (result < 0) {
            throw new UnsupportedOperationException();
        }
        return result;
    }

    /**
     * Terminates this platform resource domain.
     */
    public synchronized void terminate() {
        /*
         * After this domain is terminated, this domain instance could be
         * collected by GC, and the usage before the domain terminated should be
         * record for getTotalUsage() accounting.
         */
        saveUsageBeforeTermiante();
        int rt = handler.destroyVirtualResourceContainer(vrc);
        if (rt != 0) {
            // R1F = Fail to terminate the resource domain:\ {0}.
            throw new ResourceException(
            		com.ibm.oti.util.Msg.getString("R1F", domainName)); //$NON-NLS-1$
        }
        handler = null;

        /*
         * Return the bandwidth usage occupied by the domain to be terminated.
         */
        if (null != platformConstraint) {
            releaseBWReservation();
        }
    }

    @SuppressWarnings("boxing")
	void saveUsageBeforeTermiante() {
        synchronized (pattr) {
            this.usage = getUsage();
            Long usageOfTerminatedDomains = totalUsageConsumedByTermiantedPlatformDomains
                    .get(pattr);
            if (null == usageOfTerminatedDomains) {
                totalUsageConsumedByTermiantedPlatformDomains.put(pattr,
                        this.usage);
            } else {
                totalUsageConsumedByTermiantedPlatformDomains.put(pattr,
                        usageOfTerminatedDomains.longValue() + this.usage);
            }
        }
    }

    /**
     * Gets the resource usage consumed by all the terminated platform resource
     * domains.
     *
     * @param pattr
     *            the platform resource
     * @return the usage of the platform resource by all the terminated platform
     *         resource domains.
     */
    public static long getTotalUsageOfTermiantedDomains(
            PlatformResourceAttributes pattr) {
        Long value = totalUsageConsumedByTermiantedPlatformDomains.get(pattr);
        if (null == value) {
            return 0;
        }
        return value.longValue();
    }

    /**
     * Test whether this platform resource domain has terminated or not.
     *
     * @return true if the domain has terminated.
     */
    public boolean isTermiante() {
        return null == handler;
    }
}
