/*******************************************************************************
 * Licensed Materials - Property of IBM
 * "Restricted Materials of IBM"
 *
 *     Copyright IBM Corp. 2001, 2014 All Rights Reserved
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
package javax.rcm;

import com.ibm.rcm.resolver.ResourceConsumerResolver;
import com.ibm.rcm.resolver.ResourceConsumerResolverCenter;
import com.ibm.rcm.resource.RunningThreadsResourceAttributes;
import com.ibm.tenant.TenantContextEvent;
import com.ibm.tenant.TenantContextListener;

/**
 *	TenantContextListener used to control the running thread number for a tenant
 */
class RCMRunningThreadsTenantListener implements TenantContextListener {

    private static class Holder {
        @SuppressWarnings("synthetic-access")
        private static final RCMRunningThreadsTenantListener instance = new RCMRunningThreadsTenantListener();
    }

    private RCMRunningThreadsTenantListener() {
    }

    /**
     * @return the Singlton instance of this listener
     */
    @SuppressWarnings("synthetic-access")
    public static RCMRunningThreadsTenantListener getInstance() {
        return Holder.instance;
    }

    public void tenantContextDestroyed(TenantContextEvent event) {
        //do nothing
    }

    /**
     * Consumer one thread for the attached tenant context
     */
    public void tenantContextAttached(TenantContextEvent event) {
    	// do nothing checks are done in the AboutToAttach event
    }

    /**
     * Relinquish one thread for the attached tenant context
     */
    public void tenantContextDetached(TenantContextEvent event) {
        ResourceDomain domain = getResourceDomain(event.getTenantContext());
        if (null != domain) {
        	long rs = domain.relinquish(1);
        	if(rs != 1){
            	// R4C = Fail to release thread handle "{0}" for tenant: "{1}".
            	throw new ResourceException(com.ibm.oti.util.Msg.getString("R4C", Thread.currentThread().getName(), event.getTenantContext())); //$NON-NLS-1$
            }
        }
    }

    /*
     * Get the ResourceDomain of the consumerObject
     *
     * @param consumerObj
     * @return
     */
    private static ResourceDomain getResourceDomain(Object consumerObj) {
        ResourceConsumerResolver resolver = ResourceConsumerResolverCenter
                .getResolver();
        if (null == resolver) {
            return null;
        }

        ResourceConsumer consumer = null;
        if (null == consumerObj) {
            consumer = resolver.resolveCurrentResourceConsumer();
        } else {
            consumer = resolver.resolveResourceConsumer(consumerObj);
        }
        if (null == consumer) {
            return null;
        }
        return consumer.currentDomain(ResourceAttributesFactory
                .getInstanceByName(RunningThreadsResourceAttributes.class.getName()));
    }

	/* (non-Javadoc)
	 * @see com.ibm.tenant.TenantContextListener#tenantContextAboutToAttach(com.ibm.tenant.TenantContextEvent, java.lang.Thread)
	 */
	public void tenantContextAboutToAttach(TenantContextEvent event, Thread thread) {
        ResourceDomain domain = getResourceDomain(event.getTenantContext());
        if (null != domain) {
            long rs = domain.consume(1, false);
            if(rs != 1) {
            	// R4B = Failed to start new thread "{0}" for tenant: "{1}", thread handle exceed limits.
            	throw new ResourceException(com.ibm.oti.util.Msg.getString("R4B", thread.getName(), event.getTenantContext())); //$NON-NLS-1$
            }
        }
	}

	/* (non-Javadoc)
	 * @see com.ibm.tenant.TenantContextListener#tenantContextAttachAborted(com.ibm.tenant.TenantContextEvent, java.lang.Thread)
	 */
	public void tenantContextAttachAborted(TenantContextEvent event, Thread thread) {
        ResourceDomain domain = getResourceDomain(event.getTenantContext());
        if (null != domain) {
        	long rs = domain.relinquish(1);
        	if(rs != 1){
            	// R4C = Fail to release thread handle "{0}" for tenant: "{1}".
            	throw new ResourceException(com.ibm.oti.util.Msg.getString("R4C", thread.getName(), event.getTenantContext())); //$NON-NLS-1$
            }
        }
	}

}
