/*******************************************************************************
 * Licensed Materials - Property of IBM
 * "Restricted Materials of IBM"
 *
 *     Copyright IBM Corp. 2001, 2014 All Rights Reserved
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
package javax.rcm;

import java.util.Map;

import com.ibm.rcm.resolver.ResourceConsumerResolver;
import com.ibm.rcm.resolver.ResourceConsumerResolverCenter;
import com.ibm.rcm.resource.config.ResourceConfig;
import com.ibm.tenant.TenantContext;
import com.ibm.tenant.TenantContextEvent;
import com.ibm.tenant.TenantGlobalListener;

/**
 *
 * In AUTO_TENANT mode (-Xrcm:consumer=autotenant, see \ref consumerTypes), this
 * class listens to TenantContext's Create/Destroy event.
 *
 * For each newly created tenant, the class wrap it as a ResourceConsumer and
 * bind it to the dedicated ResourceDomain for each ResourceAttributes
 * configured by this tenant.
 *
 * For each destroyed tenant, remove it from all its ResourceDomains, and clean
 * up any controlling mechanism.
 */
class RCMTenantGlobalListener implements TenantGlobalListener {

    /**
     * The key for resource configuration in TenantConfig map.
     */
    public final static String RCM_CONFIGS_KEY = "rcm_configs"; //$NON-NLS-1$

    final static String DOMAIN_PREFIX = "#resource_domain_for_"; //$NON-NLS-1$

    /**
     * Wraps the TenantContext in ResourceConsumer and bound it to
     * ResourceDomains for each ResourceAttributes configured by this tenant.
     *
     * @param event
     *            TenantContext created event
     *
     * @see com.ibm.tenant.TenantGlobalListener#tenantGlobalCreated(com.ibm.tenant.TenantContextEvent)
     */
    public void tenantGlobalCreated(TenantContextEvent event) {

    	TenantContext tc = event.getTenantContext();

        Map<ResourceAttributes, ResourceConfig> resourceConfigs = com.ibm.tenant.internal.TenantAccessFactory.getAccess().getResourceConstraints(tc);
        if (null == resourceConfigs) {
            return;
        }

        ResourceConsumer rc = ResourceConsumer.newConsumer(tc);

        for (Map.Entry<ResourceAttributes, ResourceConfig> entry : resourceConfigs
                .entrySet()) {
            ResourceAttributes attribute = entry.getKey();
            ResourceConfig config = entry.getValue();
            if (null != config) {
                ResourceDomain domain = ResourceDomain.newDomain(attribute,
                        tc.getID() + DOMAIN_PREFIX + attribute.getName());
                config.applyToDomain(domain);
                domain.addConsumer(rc);
            }
        }
    }

    /**
     * Removes the destroyed tenant from all its ResourceDomains and clean up
     * the controlling mechanism.
     *
     * @param event
     *            TenantContext destroyed event'
     *
     * @see com.ibm.tenant.TenantGlobalListener#tenantGlobalDestroyed(com.ibm.tenant.TenantContextEvent)
     */
    public void tenantGlobalDestroyed(TenantContextEvent event) {
        TenantContext tc = event.getTenantContext();
        ResourceConsumerResolver resolver = ResourceConsumerResolverCenter
                .getResolver();
        if (null == resolver) {
            return;
        }
        ResourceConsumer rc = resolver.resolveResourceConsumer(tc);
        if (rc != null) {
            Map<ResourceAttributes, ResourceConfig> resourceConfigs = com.ibm.tenant.internal.TenantAccessFactory.getAccess().getResourceConstraints(tc);

            if (null == resourceConfigs) {
                return;
            }

            for (ResourceAttributes attribute : resourceConfigs.keySet()) {
                ResourceDomain domain = rc.currentDomain(attribute);
                domain.removeConsumer(rc);
                domain.terminate();
            }
        }
    }
}
