/*******************************************************************************
 * Licensed Materials - Property of IBM
 * "Restricted Materials of IBM"
 *
 *     Copyright IBM Corp. 2001, 2014 All Rights Reserved
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
package javax.rcm;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.lang.management.ManagementFactory;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.rcm.management.ResourceAttributesMBean;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import com.ibm.rcm.internal.util.RCMStartupOption;
import com.ibm.rcm.internal.util.ResourceSecurityHelper;
import com.ibm.rcm.resource.CpuTimeResourceAttributes;
import com.ibm.rcm.resource.FileReadResourceAttributes;
import com.ibm.rcm.resource.FileWriteResourceAttributes;
import com.ibm.rcm.resource.PlatformResourceAttributes;
import com.ibm.rcm.resource.SocketReadResourceAttributes;
import com.ibm.rcm.resource.SocketWriteResourceAttributes;
import com.ibm.rcm.resource.config.PlatformResourceConfig;
import com.ibm.rcm.resource.config.PlatformResourceConfigImpl;
import com.ibm.tenant.internal.TenantAccessFactory;

/**
 *
 * This class acts as a factory to create instances of ResourceAttributes and register them with system for management.
 *
 */
@SuppressWarnings("boxing")
public class ResourceAttributesFactory {

    private static Map<String, ResourceAttributes> defaultResources = new HashMap<String, ResourceAttributes>();

    private static Set<ResourceAttributes> registedResources = Collections
            .newSetFromMap(new ConcurrentHashMap<ResourceAttributes, Boolean>());

    private static Map<Integer, Class<? extends PlatformResourceAttributes>> defaultPlatformResources = new HashMap<Integer, Class<? extends PlatformResourceAttributes>>();

    static {
    	System.loadLibrary("j9rcm" + System.getProperty("com.ibm.oti.vm.library.version", "%%MACRO@com.ibm.oti.vm.library.version%%"));  //$NON-NLS-1$//$NON-NLS-2$ //$NON-NLS-3$

        // Add all the platform resources that native supports.
        defaultPlatformResources.put(PlatformResourceAttributes.FILE_READ_TYPE,
                FileReadResourceAttributes.class);
        defaultPlatformResources.put(
                PlatformResourceAttributes.FILE_WRITE_TYPE,
                FileWriteResourceAttributes.class);
        defaultPlatformResources.put(
                PlatformResourceAttributes.SOCKET_READ_TYPE,
                SocketReadResourceAttributes.class);
        defaultPlatformResources.put(
                PlatformResourceAttributes.SOCKET_WRITE_TYPE,
                SocketWriteResourceAttributes.class);
        defaultPlatformResources.put(PlatformResourceAttributes.CPU_TIME_TYPE,
                CpuTimeResourceAttributes.class);
        try {
            readConfiguration();
        } catch (Exception e) {
            e.printStackTrace();
            // R00 = Fail to process the resource configuration file.
            throw new ResourceException(com.ibm.oti.util.Msg.getString("R00"), e); //$NON-NLS-1$
        }

        String consumer = RCMStartupOption.getConsumer();
        if (null != consumer) {
        	/* -Xrcm:consumer=autotenant support */
            if (consumer.equals(RCMStartupOption.AUTO_TENANT)) {
            	//initialize the notification executor with root tenant wrapper
            	//so that the Notification.postConsume is ensured to be called in root tenant thread.
            	ExecutorService executor = Executors.newFixedThreadPool(InternalResourceDomain.NOTIFICATION_POOL_SIZE, TenantAccessFactory.getAccess().getRootWrappedThreadFactory(Executors.defaultThreadFactory()));
            	InternalResourceDomain.setNotificationExecutor(executor);

            	TenantAccessFactory.getAccess().addGlobalListener(new RCMTenantGlobalListener());
            /* -Xrcm:consumer=autoprocess support */
            } else if (consumer.equals(RCMStartupOption.AUTO_PROCESS)) {
            	ExecutorService executor = Executors.newFixedThreadPool(InternalResourceDomain.NOTIFICATION_POOL_SIZE,Executors.defaultThreadFactory());
            	InternalResourceDomain.setNotificationExecutor(executor);

                ResourceConsumer resourceConsumer = ResourceConsumer
                        .newConsumer(null);
                for (ResourceAttributes attr : registedResources) {
                    ResourceDomain domain = ResourceDomain.newDomain(attr,
                            "resource_domain_for_" + attr.getName()); //$NON-NLS-1$
                    /*
                     * Set the limit and reservation on the domain to be the max
                     * bandwidth. It's necessary to do such step for platform
                     * resource, otherwise there's no any resource consumption
                     * controlling on that domain.
                     */
                    if (attr instanceof PlatformResourceAttributes) {
                        PlatformResourceConfig config = null;
                        config = createPlatformResourceConfig((PlatformResourceAttributes) attr);
                        config.applyToDomain(domain);
                    }
                    domain.addConsumer(resourceConsumer);
                }
            /* -Xrcm:consumer=thread support */
            } else {
            	ExecutorService executor = Executors.newFixedThreadPool(InternalResourceDomain.NOTIFICATION_POOL_SIZE,Executors.defaultThreadFactory());
            	InternalResourceDomain.setNotificationExecutor(executor);

            }
        }
    }

    /**
     * Helper method to create a config for platform resource.
     * @param pattr the resource in question.
     * @return the config of platform resource.
     */
    private static PlatformResourceConfig createPlatformResourceConfig(
            PlatformResourceAttributes pattr) {
        PlatformResourceConfig config = new PlatformResourceConfigImpl(pattr);
        config.setBandwidthLimitByPercentage(1.0);
        config.setBandwidthReservationByPercentage(1.0);
        return config;
    }

    /**
     * Gets rcm file and reads resource config information from it.
     * @throws Exception thrown when the error happened.
     */
    private static void readConfiguration() throws Exception {
        String rcmConfigFile = System.getProperty("javax.rcm.config"); //$NON-NLS-1$
        if (null == rcmConfigFile) {
            /*
             * If the javax.rcm.config system property is not specified, the
             * default configuration file at location JAVA_HOME/lib/rcm/rcm.xml
             * is used.
             */
            rcmConfigFile = new StringBuilder()
                    .append(System.getProperty("java.home")) //$NON-NLS-1$
                    .append(File.separator).append("lib") //$NON-NLS-1$
                    .append(File.separator).append("rcm") //$NON-NLS-1$
                    .append(File.separator).append("rcm.xml").toString(); //$NON-NLS-1$
        }

        InputStream input = null;
        try {
            input = new BufferedInputStream(new FileInputStream(rcmConfigFile));
            readConfiguration(input);
        } finally {
            if (null != input) {
                input.close();
            }
        }
    }

    /**
     * Constructs managed resource instances from input stream.
     * @param in the stream to read from.
     * @throws Exception  thrown when the error happened.
     */
    private static void readConfiguration(InputStream in) throws Exception {

        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setValidating(false);
        factory.setNamespaceAware(true);
        // factory.setSchema(createSchema());

        DocumentBuilder builder = factory.newDocumentBuilder();
        builder.setErrorHandler(new ErrorHandler() {
            public void warning(SAXParseException e) throws SAXException {
                throw e;
            }

            public void error(SAXParseException e) throws SAXException {
                throw e;
            }

            public void fatalError(SAXParseException e) throws SAXException {
                throw e;
            }
        });

        Document doc = builder.parse(in);
        NodeList entries = doc.getElementsByTagName("ResourceAttributes"); //$NON-NLS-1$
        if (null == entries || 0 == entries.getLength()) {
            return;
        }
        for (int i = 0; i < entries.getLength(); i++) {
            Element entry = (Element) entries.item(i);
            String resourceName = entry.getAttribute("resourceName"); //$NON-NLS-1$
            if (null == resourceName) {
                // R3F = The "resourceName" attribute  of "ResourceAttributes" is null.
                throw new NullPointerException(
                        com.ibm.oti.util.Msg.getString("R3F")); //$NON-NLS-1$
            }
            //create resource attributes.
            Class clazz = Class.forName(resourceName);
            ResourceAttributes attributes = (ResourceAttributes) clazz.newInstance();

            Properties prop = new Properties();
            NodeList childEntries = entry
                    .getElementsByTagName("ResourceAttribute"); //$NON-NLS-1$
            if (null == childEntries || 0 == childEntries.getLength()) {
                continue;
            }
            for (int j = 0; j < childEntries.getLength(); j++) {
                Element childEntry = (Element) childEntries.item(j);
                String key = childEntry.getAttribute("name"); //$NON-NLS-1$
                String value = childEntry.getAttribute("value"); //$NON-NLS-1$
                prop.put(key, value);
            }

            attributes.configure(prop);
            defaultResources.put(resourceName, attributes);
            register(attributes);
        }
    }

    /**
     * Gets the instance of ResourceAttributes by its name.
     * @param name  the name of the resource.
     * @return an instance of the resource attributes.
     */
    public static ResourceAttributes getInstanceByName(String name) {
        if (name == null) {
            // R01 = Resource name is null.
            throw new NullPointerException(com.ibm.oti.util.Msg.getString("R01")); //$NON-NLS-1$
        }
        ResourceAttributes attr = defaultResources.get(name);

        if (null == attr) {
                // R02 = Resource "{0}" is not recognized. 
                throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString(
                        "R02", name)); //$NON-NLS-1$

        }
        return attr;
    }

    /**
     * Get  the registered resources.
     * @return a list of registered resources.
     */
    public static ResourceAttributes[] getRegistered() {
        ResourceSecurityHelper.checkQueryAllPermission();
        return registedResources
                .toArray(new ResourceAttributes[registedResources.size()]);
    }

    /**
     * Answers if the resource has been registered with system.
     * @param attributes the resource to be checked.
     * @return true if it has been in system.
     */
    private static boolean isDuplicate(ResourceAttributes attributes) {
        boolean duplicate = false;
        if (registedResources.contains(attributes)) {
            duplicate = true;
        }
        return duplicate;
    }

    /**
     * Registers a resource attributes for management.
     * @param attributes  the resource to be registered
     */
    public static void register(ResourceAttributes attributes) {
        if (null == attributes) {
            // R03 = Can not register null.
            throw new NullPointerException(com.ibm.oti.util.Msg.getString("R03")); //$NON-NLS-1$
        }
        String resourceName = attributes.getName();
        if (null == resourceName) {
            // R01 = Resource name is null.
            throw new IllegalArgumentException(
                    com.ibm.oti.util.Msg.getString("R01")); //$NON-NLS-1$
        }

        ResourceSecurityHelper.checkConfigPermission(attributes);
        validate(attributes);

        synchronized (ResourceAttributesFactory.class) {
            if (isDuplicate(attributes)) {
                // R04 = "{0}" has already registered or contained in other registered composite resources.
                throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString(
                        "R04", attributes.getName())); //$NON-NLS-1$
            }
            /*
             * If the resource attributes is platform resource, call its
             * registerRatemanagement to notify the native that the resource is
             * managed.
             */
            if (attributes instanceof PlatformResourceAttributes) {

                // Platform resource attributes cannot be registered if RCM is
                // not enabled,
                if (!RCMStartupOption.isRCMEnabled()) {
                    // R5D = Since RCM startup option is not enabled, the platform resource "{0}" is not supported, please remove it from the resource configuration file.
                    throw new IllegalStateException(com.ibm.oti.util.Msg.getString(
                            "R5D", resourceName)); //$NON-NLS-1$
                }

                // prevent registering user defined platform resource
                if (!defaultPlatformResources.values().contains(
                        attributes.getClass())) {
                    // R05 = Fail to register unrecognized platform resource:\ {0}. 
                    throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString(
                            "R05", attributes.getName())); // //$NON-NLS-1$
                }

                PlatformResourceAttributes pattr = (PlatformResourceAttributes) attributes;
                int rt =  PlatformResourceAttributesHandler.getInstance().registerManagedResource(pattr);
                if (rt != 0) {
                    // R06 = Fail to register platform resource:\ {0}.
                    throw new ResourceException(com.ibm.oti.util.Msg.getString(
                            "R06", pattr.getName())); //$NON-NLS-1$
                }

                /*
                 * Check resource handle. After the platform resource
                 * registered, the handle should be assigned to a non-zero
                 * value.
                 */
                if (0 == pattr.getHandle()) {
                    // R64 = Invalid handle of\: "{0}".
                    throw new AssertionError(com.ibm.oti.util.Msg.getString(
                            "R64",pattr.getName())); //$NON-NLS-1$
                }
                if (attributes instanceof CpuTimeResourceAttributes) {
                    // TODO use rate management period temporarily, should
                    // create a new config
                    long interval = attributes.getRateManagementPeriod() / 1000000;
                    String consumer = RCMStartupOption.getConsumer();

                    if (RCMStartupOption.AUTO_TENANT.equals(consumer)
                            || RCMStartupOption.TENANT.equals(consumer)) {
                        new TenantCPUSignalGenerator(interval).createAndStartDaemon();
                    } else if (RCMStartupOption.THREAD.equals(consumer)) {
                        new ThreadCPUSignalGenerator(interval).createAndStartDaemon();
                    } else if (RCMStartupOption.AUTO_PROCESS
                            .equals(consumer)) {
                        new ProcessCPUSignalGenerator(interval).createAndStartDaemon();
                    }
                }
            }

            /*
             * If the resource attributes is found to be a standard bean,
             * register it to the MBeanServer.
             */
            if (attributes instanceof ResourceAttributesMBean) {
                registerToMBeanServer(attributes.getName(),
                        (ResourceAttributesMBean) attributes);
            }
            registedResources.add(attributes);
        }
    }

    /**
     * Helper method to register resource with MBean server.
     * @param name used as part of object name.
     * @param mBean the bean to be registered.
     */
    private static void registerToMBeanServer(String name,
            ResourceAttributesMBean mBean) {
        try {
            MBeanServer server = ManagementFactory.getPlatformMBeanServer();
            ObjectName objName = new ObjectName(
                    "javax.rcm:type=ResourceAttributes,name=" + name); //$NON-NLS-1$
            if (!server.isRegistered(objName)) {
                server.registerMBean(mBean, objName);
            }
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Answers if the resource has been registered.
     * @param attributes the resource to be checked.
     * @return true if it has been registered.
     */
    static boolean isRegistered(ResourceAttributes attributes) {
        return registedResources.contains(attributes);
    }

    /**
     * Validate the resource attributes.
     */
    private static void validate(ResourceAttributes attributes) {
    	validateImpl(attributes);
    }

    private static void validateImpl(ResourceAttributes attributes) {
        StringBuilder errMsg = new StringBuilder(""); //$NON-NLS-1$

        if (null == attributes.getUnitName()) {
            // R07 = Unit name can not be null.
            errMsg.append(com.ibm.oti.util.Msg.getString("R07")).append("\n"); //$NON-NLS-1$ //$NON-NLS-2$
        }
        if (attributes.isUnbounded()) {
            /*
             * Unbounded resource should be neither disposable nor reservable
             * and also its total quantity must be -1.
             */
            if (attributes.isDisposable()) {
                // R08 = Disposable resource should not be unbounded.
                errMsg.append(com.ibm.oti.util.Msg.getString("R08")).append("\n"); //$NON-NLS-1$ //$NON-NLS-2$
            }
            if (attributes.isReservable()) {
                // R09 = Unbounded resource should not be reservable.
                errMsg.append(com.ibm.oti.util.Msg.getString("R09")).append("\n"); //$NON-NLS-1$ //$NON-NLS-2$
            }
            if (-1 != attributes.getTotalQuantity()) {
                // R0A = Total quantity of unbounded resource should always be -1.
                errMsg.append(com.ibm.oti.util.Msg.getString("R0A")).append("\n"); //$NON-NLS-1$ //$NON-NLS-2$
            }
        } else {
            // Total quantity of bounded resource should be a positive value.
            if (attributes.getTotalQuantity() < 0) {
                // R0B = Total quantity of the resouce should be a positive value if it is not unbounded.
                errMsg.append(com.ibm.oti.util.Msg.getString("R0B")).append("\n"); //$NON-NLS-1$ //$NON-NLS-2$
            }
        }

        if (attributes.getGranularity() < 1) {
            // R0C = Granularity should be a non-zero positive value.
            errMsg.append(com.ibm.oti.util.Msg.getString("R0C")).append("\n"); //$NON-NLS-1$ //$NON-NLS-2$
        }

        if (attributes.getMaximumMeasurementDelay() < 0
                && attributes.getMaximumMeasurementDelay() != -1) {
            // R0D = Maximum measurement delay should be a positive value or -1.
            errMsg.append(com.ibm.oti.util.Msg.getString("R0D")).append("\n"); //$NON-NLS-1$ //$NON-NLS-2$
        }

        if (attributes.getRateManagementPeriod() < 0
                && attributes.getRateManagementPeriod() != -1) {
            // R0E = Rate management period should be a positive value or -1.
            errMsg.append(com.ibm.oti.util.Msg.getString("R0E")).append("\n"); //$NON-NLS-1$ //$NON-NLS-2$
        }

        if (attributes instanceof PlatformResourceAttributes) {
            PlatformResourceAttributes pattr = (PlatformResourceAttributes) attributes;
            if (pattr.getMaxBandwidth() <= 0) {
                // R0F = Max bandwidth of platform resource should be a positive value.
                errMsg.append(com.ibm.oti.util.Msg.getString("R0F")).append("\n"); //$NON-NLS-1$ //$NON-NLS-2$
            }
            if (pattr.getRateManagementPeriod() <= 0) {
                // R10 = Rate management period of platform resource should be a positive vlaue.
                errMsg.append(com.ibm.oti.util.Msg.getString("R10")).append("\n"); //$NON-NLS-1$ //$NON-NLS-2$
            }
        }

        if (errMsg.length() > 0) {
            // R11 = Fail to register the invalid resource "{0}". The reason is:\ {1}.
            throw new IllegalArgumentException(
                    com.ibm.oti.util.Msg.getString(
                            "R11", new Object[] { attributes.getName(), errMsg.toString() })); //$NON-NLS-1$
        }
    }

    // private constructor
    private ResourceAttributesFactory() {
    };

}
