/*******************************************************************************
 * Licensed Materials - Property of IBM
 * "Restricted Materials of IBM"
 *
 *     Copyright IBM Corp. 2001, 2014 All Rights Reserved
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
package javax.rcm;

import java.util.concurrent.ConcurrentHashMap;

import com.ibm.rcm.internal.util.ResourceSecurityHelper;
import com.ibm.rcm.resolver.ResourceConsumerResolver;
import com.ibm.rcm.resolver.ResourceConsumerResolverCenter;

/**
 * This class is representing a executable entity such likes a Thread or a tenant whose resource consumption
 * is controlled by resource management mechanism.
 *
 */
public final class ResourceConsumer {
    private final Object consumer;

    private ConcurrentHashMap<ResourceAttributes, ResourceDomain> boundDomains = new ConcurrentHashMap<ResourceAttributes, ResourceDomain>();

    /**
     * Gets all resource consumers in the system.
     * @return array of current resource consumers.
     */
    public static ResourceConsumer[] currentConsumers() {
        ResourceSecurityHelper.checkQueryAllPermission();

        ResourceConsumerResolver resolver = ResourceConsumerResolverCenter
                .getResolver();
        if (null == resolver) {
            return new ResourceConsumer[] {};
        }
        ResourceConsumer consumer = resolver.resolveCurrentResourceConsumer();
        if (null == consumer) {
            return new ResourceConsumer[] {};
        }
        return new ResourceConsumer[] { consumer };
    }

    /**
     * Creates a new resource consumer. The exception UnsupportedOperationException is thrown if the consumer type is not supported
     * by system.
     * @param consumer the actual consume object wrapped inside ResourceConsumer instance.
     * @return a ResourceConsumer wrapper around the actual consumer.
     */
    public static ResourceConsumer newConsumer(Object consumer) {
        ResourceSecurityHelper.checkControlAllPermission();
        if (!checkConsumer(consumer)) {
            // R20 = Unsupported resource consumer type:\ {0}.
            if (null == consumer) {
                throw new UnsupportedOperationException(com.ibm.oti.util.Msg.getString(
                        "R20", "null")); //$NON-NLS-1$ //$NON-NLS-2$
            } else {
                throw new UnsupportedOperationException(com.ibm.oti.util.Msg.getString(
                        "R20", consumer.getClass().getName())); //$NON-NLS-1$
            }
        }

        ResourceConsumerResolver resolver = ResourceConsumerResolverCenter
                .getResolver();
        ResourceConsumer rc = resolver.resolveResourceConsumer(consumer);
        if (null == rc) {
            rc = new ResourceConsumer(consumer);
            resolver.register(rc);
        }

        return rc;
    }

    /**
     * Check if the type of the consumer is supported or not.
     */
    private static boolean checkConsumer(Object consumer) {
        return ResourceConsumerResolverCenter.isConsumerTypeSupported(consumer);
    }

    private ResourceConsumer(Object consumer) {
        this.consumer = consumer;
    }

    /**
     * Queries the current domain for the specified resource.
     * @param resourceAttributes  the the resource in question.
     * @return resource domain for this resource or null if it doens't exit.
     */
    public ResourceDomain currentDomain(ResourceAttributes resourceAttributes) {
        ResourceSecurityHelper.checkQueryAllPermission();
        return currentDomainImpl(resourceAttributes);
    }

    ResourceDomain currentDomainImpl(ResourceAttributes resourceAttributes) {
        return boundDomains.get(resourceAttributes);
    }

    /**
     * Tells all resource domains to which this consumer is bound.
     * @return all the resource domains to which this consumer is bound
     */
    public ResourceDomain[] currentDomains() {
        ResourceSecurityHelper.checkQueryAllPermission();
        ResourceDomain[] res = new ResourceDomain[boundDomains.size()];
        return boundDomains.values().toArray(res);
    }
    /**
     * Test if this instance of ResourceConsumer is equal to another.
     * @param obj another instance of ResourceConsumer.
     * @return return true if it is equal to another.
     */
    public boolean equals(Object obj) {
        if (null == obj || !(obj instanceof ResourceConsumer)) {
            return false;
        }
        ResourceConsumer rc = (ResourceConsumer) obj;
        if (null == consumer) {
            return null == rc.getConsumerImpl();
        }
        return consumer.equals(rc.getConsumerImpl());
    }
    /**
     * Gets the actual consumer object wrapped inside this instance.
     * @return the actual consume object.
     */
    public Object getConsumer() {
        ResourceSecurityHelper.checkQueryAllPermission();
        return getConsumerImpl();
    }

    Object getConsumerImpl() {
        return this.consumer;
    }

    /**
     * Overrides the hashCode in java.lang.Object.
     * @return the hash code of this instance.
     */
    public int hashCode() {
        if (null == consumer) {
            return 0;
        }
        return consumer.hashCode();
    }

    void boundToResourceDomain(ResourceDomain domain) {
        boundDomains.putIfAbsent(domain.getResourceAttributes(), domain);
    }

    void unboundFromResourceDomain(ResourceDomain domain) {
        boundDomains.remove(domain.getResourceAttributes());
    }
}
