/*******************************************************************************
 * Licensed Materials - Property of IBM
 * "Restricted Materials of IBM"
 *
 *     Copyright IBM Corp. 2001, 2014 All Rights Reserved
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
package javax.rcm;

import java.lang.management.ManagementFactory;
import java.util.concurrent.CopyOnWriteArrayList;

import javax.management.MBeanServer;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;
import javax.management.StandardMBean;
import javax.rcm.management.ResourceDomainMBean;

import com.ibm.rcm.internal.util.ResourceSecurityHelper;

/**
 * The resource domain is to:
 * bind a group of resource consumers to the resource  which is managed by this domain instance.
 * and impose the consumption policy on this consumption group.
 *
 *
 */
public final class ResourceDomain {

    private ResourceAttributes resourceAttributes;

    private String name;

    private InternalResourceDomain internal;

    /**
     * All the consumers bounded to this domain.
     */
    private CopyOnWriteArrayList<ResourceConsumer> consumers = new CopyOnWriteArrayList<ResourceConsumer>();

    /**
     * Creates a resource domain for the specified resource.
     * @param resourceAttributes the resource which this domain manages
     * @param domainName unique name for the new domain
     * @return a new domain for the resource.
     */
    public static ResourceDomain newDomain(
            ResourceAttributes resourceAttributes, String domainName) {
        ResourceSecurityHelper.checkControlAllPermission();

        if (null == resourceAttributes || null == domainName) {
            // R26 = Resource attribute or domain name can not be null.
            throw new NullPointerException(com.ibm.oti.util.Msg.getString("R26")); //$NON-NLS-1$
        }

        if (!ResourceAttributesFactory.isRegistered(resourceAttributes)) {
            // R27 = Fail to create domain "{0}". Resource "{1}" hasn't be registered.
            throw new IllegalStateException(com.ibm.oti.util.Msg.getString("R27", //$NON-NLS-1$
                    new String[] { domainName, resourceAttributes.getName() }));
        }

        ResourceDomain domain = new ResourceDomain(resourceAttributes,
                domainName);

        /*
         * Create an MBean for the created resource domain, and register it to
         * the MBean server.
         */
        registerMBean(domain);

        return domain;
    }

    private ResourceDomain(ResourceAttributes resourceAttributes, String name) {
        this.resourceAttributes = resourceAttributes;
        this.name = name;
        this.internal = new InternalResourceDomain(this);
    }

    /**
     * Adds a constraint to this domain.
     * @param constraint the constraint to be added.
     */
    public void addConstraint(Constraint constraint) {
        ResourceSecurityHelper.checkControlPermission(resourceAttributes);
        internal.addConstraint(constraint);
    }
    /**
     * Binds a resource consumer to this domain.
     * @param consumer the consumer to be binded.
     */
    public void addConsumer(ResourceConsumer consumer) {
        ResourceSecurityHelper.checkControlPermission(resourceAttributes);

        if (isTerminated()) {
            // R36 = Resource domain is terminated.
            throw new IllegalStateException(com.ibm.oti.util.Msg.getString("R36", //$NON-NLS-1$
                    name));
        }

        if (null == consumer) {
            throw new NullPointerException();
        }

        synchronized (this) {
            ResourceDomain domainOfSameResource = consumer
                    .currentDomainImpl(resourceAttributes);

            if (consumers.contains(consumer) || null != domainOfSameResource) {
                // R2C = Consumer already bound to a domain for resource "{0}".
                throw new IllegalStateException(com.ibm.oti.util.Msg.getString(
                        "R2C", resourceAttributes)); //$NON-NLS-1$
            }
            internal.addConsumer(consumer.getConsumerImpl());
            consumer.boundToResourceDomain(this);
            consumers.add(consumer);
        }
    }

    /**
     * Adds a notification to this domain.
     * @param notification the notification to be binded.
     */
    public void addNotification(Notification notification) {
        ResourceSecurityHelper.checkControlPermission(resourceAttributes);
        internal.addNotification(notification);
    }
    /**
     * Determines if a specified quantity of the resource can be consumed.
     * @param quantity  requested quantity of the resource.
     * @param allowPartialGrant indicates if the returned value can be less than requested
     * @return the quantity of consumption units of resource which was actually granted.
     */
    public long consume(long quantity, boolean allowPartialGrant) {
        ResourceSecurityHelper.checkConsumePermission(resourceAttributes);
        return internal.consume(quantity, allowPartialGrant);
    }
    /**
     * Gets the name of resource domain.
     * @return the name of domain.
     */
    public String getName() {
        return this.name;
    }

    /**
     * Gets the resource which is bound to this domain instance.
     * @return the resource managed by this resource domain
     */
    public ResourceAttributes getResourceAttributes() {
        ResourceSecurityHelper.checkQueryPermission(resourceAttributes);
        return this.resourceAttributes;
    }
    /**
     * Gets the resource consumers bound to this domain instance.
     * @return the resource consumers bound to this domain.
     */
    public ResourceConsumer[] getResourceConsumers() {
        ResourceSecurityHelper.checkQueryPermission(resourceAttributes);
        return consumers.toArray(new ResourceConsumer[consumers.size()]);
    }
    /**
     * Gets the total reserved amount of the resource which is controlled by this domain instnace.
     * @return the total reserved amount of the resource.
     */
    public long getTotalReservation() {
        ResourceSecurityHelper.checkQueryPermission(resourceAttributes);
        return internal.getTotalReservation();
    }
    /**
     * Gets the reservation associated with this instance.
     * @return the reservation associated with this resource domain.
     */
    public long getReservation() {
        ResourceSecurityHelper.checkQueryPermission(resourceAttributes);
        return internal.getReservation();
    }
    /**
     * Get the total usage of resource in all resource domains.
     * @return the total usage of resource.
     */
    public long getTotalUsage() {
        ResourceSecurityHelper.checkQueryPermission(resourceAttributes);
        return internal.getTotalUsage();
    }
   /**
    * Gets the usage of the resource in this resource domain.
    * @return the usage of the resource.
    */
    public long getUsage() {
        ResourceSecurityHelper.checkQueryPermission(resourceAttributes);
        return internal.getUsage();
    }
    /**
     * Gets the total rate of usage of the resource in all resource domains.
     * @return the total rate of usage of the resource.
     */
    public long getTotalUsageRate() {
        ResourceSecurityHelper.checkQueryPermission(resourceAttributes);
        return internal.getTotalUsageRate();
    }
    /**
     * Gets the total rate of usage of the resource in this resource domain.
     * @return the rate of usage of the resource.
     */
    public long getUsageRate() {
        ResourceSecurityHelper.checkQueryPermission(resourceAttributes);
        return internal.getUsageRate();
    }
    /**
     * Tests if the domain is in the terminated state.
     * @return true it is in terminated state.
     */
    public boolean isTerminated() {
        return internal.isTerminated();
    }
    /**
     * Relinquishes a specified quantity of the resource which is no longer in use to this domain.
     * @param quantity the amount of resource that is being released.
     * @return the quantity of the resource that was actually released.
     */
    public long relinquish(long quantity) {
        ResourceSecurityHelper.checkConsumePermission(resourceAttributes);
        return internal.relinquish(quantity);
    }
    /**
     * Removes the constraint from this resource domain.
     * @param constraint the constraint to be removed.
     */
    public void removeConstraint(Constraint constraint) {
        ResourceSecurityHelper.checkControlPermission(resourceAttributes);
        internal.removeConstraint(constraint);
    }
    /**
     * Removes the consumer from this resource domain.
     * @param resourceConsumer the consumer to be removed.
     * @return true if resourceConsumer was previously bound to this domain.
     */
    public boolean removeConsumer(ResourceConsumer resourceConsumer) {
        ResourceSecurityHelper.checkControlPermission(resourceAttributes);

        if (isTerminated()) {
            // R36 = Resource domain is terminated.
            throw new IllegalStateException(com.ibm.oti.util.Msg.getString("R36", //$NON-NLS-1$
                    name));
        }

        if (null == resourceConsumer) {
            throw new NullPointerException();
        }

        synchronized (this) {
            if (!consumers.contains(resourceConsumer)) {
                return false;
            }
            internal.removeConsumer(resourceConsumer.getConsumerImpl());
            consumers.remove(resourceConsumer);
            resourceConsumer.unboundFromResourceDomain(this);
            return true;
        }
    }
    /**
     * Removes the notification from this resource domain.
     * @param notification the notification to be removed.
     */
    public void removeNotification(Notification notification) {
        ResourceSecurityHelper.checkControlPermission(resourceAttributes);
        internal.removeNotification(notification);
    }
    /**
     * Sets a resource usage reservation on this domain.
     * @param newValue the new reservation for resource consumption.
     */
    public void setReservation(long newValue) {
        ResourceSecurityHelper.checkControlPermission(resourceAttributes);
        internal.setReservation(newValue);
    }
    /**
     * Terminates this resource domain.
     */
    public void terminate() {
        ResourceSecurityHelper.checkControlPermission(resourceAttributes);

        if (isTerminated()) {
            return;
        }

        // Check if any consumer is still bound to this domain.
        if (!consumers.isEmpty()) {
            // R35 = Can not terminate the resource domain now as there are still consumers bound to it.
            throw new IllegalStateException(com.ibm.oti.util.Msg.getString("R35")); //$NON-NLS-1$
        }

        internal.terminate();

        /**
         * Unregister the resource domain MBean after the domain is terminated,
         * so that the domain instance can be collected by GC.
         */
        unregisterMBean(this);
    }

    private static ObjectName createObjectName(ResourceDomain domain)
            throws MalformedObjectNameException {
    	return new ObjectName("javax.rcm:type=ResourceDomain,name=\"" //$NON-NLS-1$
                    + domain.getName() + "\""); //$NON-NLS-1$

    }

    static void registerMBean(ResourceDomain domain) {
        MBeanServer server = ManagementFactory.getPlatformMBeanServer();
        if (server != null) {
            try {
                ObjectName objName = createObjectName(domain);
                StandardMBean mBean = new StandardMBean(new ResourceDomainMBeanImpl(domain), ResourceDomainMBean.class);
                server.registerMBean(mBean, objName);
            } catch (Exception e) {
                throw new RuntimeException(e);
            }
        }
    }

    static void unregisterMBean(ResourceDomain domain) {
        MBeanServer server = ManagementFactory.getPlatformMBeanServer();
        if (server != null) {
            ObjectName objName;
            try {
                objName = createObjectName(domain);
                server.unregisterMBean(objName);
            } catch (Exception e) {
                throw new RuntimeException(e);
            }
        }
    }

    long applyConstraints(long current, long proposed) {
        return internal.applyConstraints(current, proposed);
    }

    void sendNotifications(long perviousUsage, long currentUsage) {
        internal.sendNotifications(perviousUsage, currentUsage);
    }

    public int hashCode() {
        return name.hashCode();
    }
    /**
     * Returns a string description of this resource domain.
     */
    public String toString() {
        return "Resource domain [" + super.toString() + "]" + " - " //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
                + resourceAttributes.getName();
    }
}
