/*******************************************************************************
 * Licensed Materials - Property of IBM
 * "Restricted Materials of IBM"
 *
 *     Copyright IBM Corp. 2001, 2014 All Rights Reserved
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
package javax.rcm;

import java.security.BasicPermission;
import java.security.Permission;
import java.security.PermissionCollection;

/**
 * permission class is used to implement a security policy that allows
 * only trusted code to do privileged actions via the Resource Management API.
 *
 */
public final class ResourcePermission extends BasicPermission {

    /**
     * Serial version UID.
     */
    private static final long serialVersionUID = -8848936956631501401L;

    // List of actions permitted for resource permission in order.
    private static final String[] actionList = { "query", "control", //$NON-NLS-1$ //$NON-NLS-2$
            "configure", "consume" };  //$NON-NLS-1$//$NON-NLS-2$

    /*
     * The numeric representation of action list for implies() to check if one
     * action list is the subset of another.
     */
    transient int mask = -1;

    private String actions;

    /**
     * Constructs a new resource permission for a resource.
     * @param resource the name of the resource being managed.
     * @param actions the actions of permission.
     */
    public ResourcePermission(String resource, String actions) {
        super(resource);
        init(resource, actions);
    }

    private void init(String resource, String theActions) {
        if (null == theActions || "".equals(theActions)) { //$NON-NLS-1$
            // R22 = Actions can not be null or empty.
            throw new NullPointerException(com.ibm.oti.util.Msg.getString("R22")); //$NON-NLS-1$
        }
        this.mask = getMask(theActions);
        this.actions = toActionString();
    }

    private int getMask(String theActions) {
        int theMask = 0, head = 0, tail = 0;
        do {
            tail = theActions.indexOf(",", head); //$NON-NLS-1$
            String action = tail > 0 ? theActions.substring(head, tail).trim()
                    : theActions.substring(head).trim();
            if (action.equals(actionList[0])) {
                theMask |= 8;
            } else if (action.equals(actionList[1])) {
                theMask |= 4;
            } else if (action.equals(actionList[2])) {
                theMask |= 2;
            } else if (action.equals(actionList[3])) {
                theMask |= 1;
            } else {
                // R23 = Invalid actions. The value should be control, query, consume and configure.
                throw new IllegalArgumentException(com.ibm.oti.util.Msg.getString(
                        "R23", action)); //$NON-NLS-1$
            }
            head = tail + 1;
        } while (tail > 0);
        return theMask;
    }

    private String toActionString() {
        StringBuffer buffer = new StringBuffer(""); //$NON-NLS-1$
        // test from the leftmost position
        int testBitMask = 1 << (actionList.length - 1);

        for (int i = 0; i < actionList.length; i++) {
            if ((mask & testBitMask) != 0) {
                buffer.append(actionList[i]);
                buffer.append(","); //$NON-NLS-1$
            }
            testBitMask = testBitMask >> 1;
        }
        // remove the last comma
        if (buffer.length() > 0) {
            buffer.deleteCharAt(buffer.length() - 1);
        }
        return buffer.toString();
    }

    /**
     * Returns a new PermissionCollection in which to place ResourcePermission
     * objects.
     *
     * @return A new PermissionCollection object suitable for storing
     *         ResourcePermission objects.
     */
    @Override
    public PermissionCollection newPermissionCollection() {
        return new ResourcePermissionCollection();
    }

    @Override
    public boolean equals(Object obj) {
        if (obj instanceof ResourcePermission) {
            ResourcePermission rp = (ResourcePermission) obj;
            if (rp.actions != actions) {
                if (null == rp.actions || !rp.actions.equals(actions)) {
                    return false;
                }
            }
            return rp.getName().equals(this.getName());
        }
        return false;
    }

    /**
     * return ordered actions split by comma.
     */
    @Override
    public String getActions() {
        return this.actions;
    }

    @Override
    public int hashCode() {
        return getName().hashCode();
    }

    @Override
    public boolean implies(Permission permission) {
        int matchedMask = impliesMask(permission);
        return matchedMask != 0
                && matchedMask == ((ResourcePermission) permission).mask;
    }

    int impliesMask(Permission permission) {
        if (null == permission || !(permission instanceof ResourcePermission)) {
            return 0;
        }

        if (!super.implies(permission)) {
            return 0;
        }

        ResourcePermission rp = (ResourcePermission) permission;

        int theMask = this.mask;
        if ((theMask & (1 << 2)) != 0) {
            theMask = theMask | (1 << 3);
        }
        int matchedMask = theMask & rp.mask;

        return matchedMask;
    }
}
